/**
  *  \file arch/unix/cmdargs.cc
  *  \brief Platform-specific option parsing for Unix
  *
  *  These are in a separate file due to linking dependencies.
  *  Putting this in platform.cc would reference util/cmdline.cc
  *  each time it is used (backwards dependency in linking).
  *  This is not a problem when cmdargs.cc is separate: a file
  *  which calls Platform::initOptions most likely also calls
  *  the cmdline.cc module, hence the required symbols are available.
  */

#define _POSIX_SOURCE           // putenv
#define _BSD_SOURCE             // setenv
#define _XOPEN_SOURCE           // putenv

#include <stdlib.h>
#include <string.h>
#include "arch/unix/platform.h"
#include "defs.h"
#include "util/nls.h"
#include "util/cmdline.h"
#include "cpluslib/misc.h"
#ifdef HAVE_SDL
#  include "SDL.h"
#endif

/** Handle -display option. Unfortunately, SDL has no better way to do
    this. */
static void
optSetDisplay(StandardCommandLineParser& p, const StandardCommandLineParser::Option&)
{
#if defined(HAVE_PUTENV)
    /* POSIX way. When platform doesn't have putenv, SDL emulates it.
       Some library versions do not copy the string (which seems to be
       a requirement of SUSv2), hence we must copy it. This will cause
       a memory leak each time "-display" is used, let's hope this
       isn't serious :) */
    char* value = new char[sizeof("DISPLAY=") + strlen(p.getArg())];
    strcpy(value, "DISPLAY=");
    strcat(value, p.getArg());
    putenv(value);
#elif defined(HAVE_SETENV)
    /* BSD way */
    setenv("DISPLAY", p.getArg(), 1);
#endif
}

static const StandardCommandLineParser::Option options[] = {
    { "display", optSetDisplay, 0, true, N_(" -display DPY       X display to use (if applicable)") }
};

void
Platform::initOptions(StandardCommandLineParser& par)
{
    par.addOptionSet(::options, countof(::options));
}
