/**
  *  \file client/actions/gaexport.cc
  *  \brief Global Actions: Export
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <memory>
#include "client/actions/gaexport.h"
#include "client/widgets/expfieldlist.h"
#include "client/widgets/expformat.h"
#include "client/widgets/helpwidget.h"
#include "cpluslib/format.h"
#include "game/error.h"
#include "game/planet.h"
#include "game/pref.h"
#include "game/ship.h"
#include "gfx/palette.h"
#include "int/export.h"
#include "int/if/planetif.h"
#include "int/if/shipif.h"
#include "int/interror.h"
#include "io/glob.h"
#include "ui/dialogs/filedlg.h"
#include "ui/msgbox.h"
#include "ui/widgets/button.h"
#include "ui/widgets/framegroup.h"
#include "ui/widgets/statictext.h"
#include "util/answer.h"
#include "util/nls.h"
#include "ui/layout.h"

namespace {
    class WExportParameterDialog : public UIWindow {
        IntExportFieldList& fields;
        IntContext* ctx;
        WExportFormatControl* format_control;
        AnswerProvider& filter;
        ObjectHolder h;
        void init();
        void onOK();
     public:
        WExportParameterDialog(IntExportFieldList& fields, IntContext* ctx, AnswerProvider& filter);
    };
}

/************************* WExportParameterDialog ************************/

WExportParameterDialog::WExportParameterDialog(IntExportFieldList& fields,
                                               IntContext* ctx,
                                               AnswerProvider& filter)
    : UIWindow(_("Export"), BLUE_WINDOW, COLOR_GRAY),
      fields(fields),
      ctx(ctx),
      filter(filter)
{
    init();
}

void
WExportParameterDialog::init()
{
    // VBox
    //   UIFrameGroup/HBox
    //     WExportFieldList
    //     UIScrollbar
    //   HBox
    //     UIButton "Ins"
    //     UIButton "Del"
    //     UISpacer
    //  [fileformatselect]
    //   HBox
    //     "OK"
    //     "Cancel"
    //     UISpacer
    //     "Help"

    setLayout(UIVBoxLayout::instance5);

    UIFrameGroup& g1 = h.add(new UIFrameGroup(UIHBoxLayout::instance0));
    UIGroup&      g2 = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup&      g4 = h.add(new UIGroup(UIHBoxLayout::instance5));

    // Field list
    WExportFieldList& xfl = h.add(new WExportFieldList(fields, ctx));
    g1.add(xfl);
    g1.add(h.add(new UIScrollbar(xfl)));
    add(g1);

    // Field list buttons
    UIButton* btn = &h.add(new UIButton(_("Ins"), SDLK_INSERT));
    btn->setState(UIButton::bf_SmallSize, true);
    g2.add(*btn);
    g2.add(h.add(new UIStaticText(_("Add..."), UIStaticText::StaticNormal)));

    btn = &h.add(new UIButton(_("Del"), SDLK_INSERT));
    btn->setState(UIButton::bf_SmallSize, true);
    g2.add(*btn);
    g2.add(h.add(new UIStaticText(_("Remove"), UIStaticText::StaticNormal)));

    g2.add(h.add(new UISpacer()));

    g2.add(h.add(new UIStaticText(_("Width"), UIStaticText::StaticNormal)));
    btn = &h.add(new UIButton("+", '+'));
    btn->setState(UIButton::bf_SmallSize, true);
    g2.add(*btn);
    btn = &h.add(new UIButton("-", '-'));
    btn->setState(UIButton::bf_SmallSize, true);
    g2.add(*btn);

    add(g2);

    // File format select
    format_control = &h.add(new WExportFormatControl());
    add(*format_control);

    // Dialog buttons
    g4.add(h.add(new UIButton(1, _("OK"), SDLK_RETURN)).addCall(this, &WExportParameterDialog::onOK));
    g4.add(h.add(new UIButton(0, _("Cancel"), SDLK_ESCAPE)).addStop());
    g4.add(h.add(new UISpacer()));
    g4.add(h.add(new UIButton(_("Help"), 'h')));
    add(g4);

    add(h.add(new WHelpWidget("pcc2:export")));

    pack();
}

/** OK button. */
void
WExportParameterDialog::onOK()
{
    // Must have some fields
    if (fields.size() == 0) {
        messageBox(_("Please select some fields for output."), _("Export"));
        return;
    }

    // Select output file name
    string_t ext = format_control->getRecommendedFileExtension();
    UIFileChooserDialog dlg(_("Export"),
                            Directory::openDir(Directory::getWorkingDirectory()),
                            FileNamePattern::getAllFilesWithExtensionPattern(ext));
    ui_root->centerWidget(dlg);
    if (ui_root->runModal(dlg)) {
        string_t name = appendFileNameExtension(dlg.getResultFile(), ext, false);
        try {
            Ptr<Stream> s = Stream::openFile(name, Stream::C_CREATE);
            format_control->doExport(*s, ctx, filter, fields);
            stop(1);
        }
        catch (IntError& e) {
            messageBox(format(_("Error during export: %s"), e.what()),
                       _("Export"));
        }
        catch (GError& e) {
            messageBox(format(_("Error during export: %s"), e.getUserError()),
                       _("Export"));
        }
        catch (FileProblemException& e) {
            messageBox(format(_("Unable to create %s: %s"), e.getFileName(), e.what()),
                       _("Export"));
        }
        catch (std::runtime_error& e) {
            messageBox(format(_("Unable to create %s: %s"), name, e.what()),
                       _("Export"));
        }
    }
}

/************************** WGlobalExportAction **************************/

/** Export action. The execute function just collects objects, the
    actual work is done in finish().

    The Ids of the objects to export are collected in %objects.
    The %state tracks which kind of object we're collecting, and
    detects problems. When objects have successfully been collected,
    this object also acts as an AnswerProvider used for filtering
    the export. */
class WGlobalExportAction::Executer
    : public WGlobalActionExecuter,
      private AnswerProvider
{
    enum {
        Neutral,
        CollectPlanets,
        CollectShips,
        Failure
    } state;

    enum {
        MAX_OBJS = NUM_SHIPS > NUM_PLANETS ? NUM_SHIPS : NUM_PLANETS,
        WORDS = MAX_OBJS / 32 + 1
    };

    uint32_t objects[WORDS];

 public:
    Executer()
        : state(Neutral)
        {
            std::fill(objects, objects+WORDS, uint32_t(0));
        }

    // WGlobalActionExecuter:
    bool execute(GObject& obj, bool /*lock*/);
    void finish(int obj, int success);

    // AnswerProvider:
    bool ask(int questionId, string_t question);
};

WGlobalActionExecuter*
WGlobalExportAction::prepare()
{
    return new Executer();
}

bool
WGlobalExportAction::Executer::execute(GObject& obj, bool /*lock*/)
{
    GPlanet* pl = dynamic_cast<GPlanet*>(&obj);
    GShip* sh = dynamic_cast<GShip*>(&obj);

    // Ignore anything but ships or planets.
    if (pl != 0 || sh != 0) {
        // If still neutral, infer type.
        if (state == Neutral) {
            if (pl != 0) {
                state = CollectPlanets;
            } else {
                state = CollectShips;
            }
        }

        // Make sure we only receive objects of one type.
        if (((state == CollectShips && sh != 0)
             || (state == CollectPlanets && pl != 0))
            && obj.getId() > 0
            && obj.getId() <= MAX_OBJS)
        {
            uint32_t slot = obj.getId();
            objects[slot / 32] |= 1 << (slot % 32);
        } else {
            state = Failure;
        }
    }

    return true;
}

void
WGlobalExportAction::Executer::finish(int /*obj*/, int /*success*/)
{
    // Handle failure modes
    if (state == Neutral) {
        messageBox(_("No units selected for export."), _("Export"));
        return;
    }
    if (state == Failure) {
        messageBox(_("Your selection includes ships and planets. "
                     "PCC cannot export ships and planets at the same time. "
                     "Please use the \"1\" and \"2\" buttons to select only one type."),
                   _("Export"));
        return;
    }

    // Export field list
    ConfigStringOption& opt = (state == CollectPlanets
                               ? getUserPreferences().ExportPlanetFields
                               : getUserPreferences().ExportShipFields);
    IntExportFieldList fields;
    try {
        fields.addList(opt());
    }
    catch (...)
    { }

    // Export object type
    std::auto_ptr<IntContext> type(state == CollectPlanets
                                   ? IFPlanetMake()
                                   : IFShipMake());
    if (!type.get()) {
        // This should not happen, but better be safe.
        messageBox(_("No units selected for export."), _("Export"));
        return;
    }

    // Dialog
    WExportParameterDialog dialog(fields, type.get(), *this);
    ui_root->centerWidget(dialog);
    if (ui_root->runModal(dialog)) {
        // The dialog will have performed the export. However, we still
        // have to remember the field list for next time.
        opt.set(fields.toString());

        // Information. We only get here when export really succeeded;
        // exceptions are caught by onOK and keep the dialog open.
        messageBox(_("Export succeeded."), _("Export"));
    }
}

bool
WGlobalExportAction::Executer::ask(int questionId, string_t /*question*/)
{
    return questionId > 0
        && questionId <= MAX_OBJS
        && (objects[questionId / 32] & (1 << (questionId % 32))) != 0;
}
