/**
  *  \file int/value.h
  *  \brief Interpreter: Values
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_INT_VALUE_H
#define PCC_V2_INT_VALUE_H

#include <vector>
#include "cpluslib/string.h"
#include "cpluslib/types.h"

class Stream;
struct TStructureParser;

/** Serialized form of a value. This is the same structure as used in PCC 1.x.
    A serialized data segment consists of a sequence of these 48-bit tag nodes,
    followed by (optional) auxiliary data. The tag nodes contain
    - 32-bit integer or bool
    - string tag (actual string data in auxiliary data field)
    - 48-bit floating point values (Turbo Pascal REAL)
    - 32-bit floating point values (IEEE single precision)

    Turbo Pascal has an entirely-software emulated floating point type REAL. Those
    have the convenient property of having an effective value of 0.0 if their first
    byte, the exponent, is zero. All the other bytes are irrelevant in this case
    (except for the 47th bit, the sign bit, which distinguishes positive and negative
    zeroes, to be precise). PCC 1.x therefore makes sure that actual floating-point
    zeroes are stored as all-48-bits-zero, and uses the "non-standard" zeroes
    to store tag/value pairs. The tags thus need to have zero in the lower 8 bits.

    Non-scalar values require additional data in the auxiliary data field. This data
    follows the tag nodes in sequence. This has the drawback that random access is
    not possible, just sequential reading, and it also means that there is no way
    to skip over data added by potential future tags, i.e. non-scalars after those
    future values are lost. PCC 1.x, in anticipation of PCC2, implements a little
    forward compatibility by reading 32-bit floats and long strings, although it
    never writes them.

    Although 32-bit floats were explicitly intended to be used with PCC2, it turned
    out that they have too little precision for our needs, so we don't use them as
    well. PCC2 serializes floats as 48-bit REAL. */
struct IntTagNode {
    uint16 tag;
    uint32 value;

    enum {
        Tag_FPZero     = 0x0000,       ///< Floating-point zero.
        Tag_Empty      = 0x0100,       ///< EMPTY (null, unassigned, ...).
        Tag_Integer    = 0x0200,       ///< Integer. value is actual value.
        Tag_Boolean    = 0x0300,       ///< Boolean. value is 0 or 1.
        Tag_String     = 0x0400,       ///< String. value is flag: 0=empty string, 1=Pascal string follows in auxiliary data.
        Tag_32bitFP    = 0x0500,       ///< 32-bit floating point. value is IEEE single precision value. Supported since PCC 1.0.9.
        Tag_LongString = 0x0600,       ///< String. value is number of bytes following in auxiliary data. Supported since PCC 1.0.18.
        Tag_BCO        = 0x0700,       ///< Bytecode object. Only valid within VM files. value is the BCO's id.
        Tag_Array      = 0x0800,       ///< Array. Only valid within VM files. value is array Id.
        Tag_Blob       = 0x0900,       ///< Byte string. Works like Tag_LongString, but content is byte string, not text, for file I/O.
        Tag_FileHandle = 0x0A00,       ///< File handle. Works like Tag_Integer, but content is file handle ("#10").
        Tag_Hash       = 0x0B00,       ///< Hash. Only valid within VM files. value is hash Id.
        Tag_Struct     = 0x0C00,       ///< Structure. Only valid within VM files. value is struct Id.
        Tag_StructType = 0x0D00        ///< Structure type. Only valid within VM files. value is struct type Id.
    };

    // Serialisation information
    static const TStructureParser parser[];
    enum {
        size  = 6,
        count = 2
    };
};

/** Generic value. This is the base class for all values used in the interpreter.
    Note that all places that use an IntValue* also allow it to be null, representing
    the CCScript EMPTY (null) value used for unknown properties. */
class IntValue {
 public:
    /** Destructor. */
    virtual ~IntValue();

    /** Convert this value to a string.
        \param readable true=make it readable to the lexer; false=program-controlled interpolation */
    virtual string_t toString(bool readable) = 0;

    /** Store this value to a stream.
        \param sv [out] Header (tag) information goes here
        \param aux [out] Auxiliary data stored here, if required
        \see IntTagNode */
    virtual void store(IntTagNode& sv, Stream& aux) = 0;

    /** Clone this value. Returns a duplicate of this value. The caller assumes responsibility for
        the new object. */
    virtual IntValue* clone() = 0;

    /** Load a value.
        \param sv  [in] Header (tag) information
        \param aux [in] Auxiliary data stream, auxiliary data read here if required
        \return newly-allocated IntValue descendant (or null)
        \see IntTagNode */
    static IntValue* load(const IntTagNode& sv, Stream& aux);

    /** Clone a value. Same as p->clone(), but allows p to be 0. */
    static IntValue* cloneOf(IntValue* p);

    /** Convert a value into a string. Same as p->toString(readable), but allows p to be 0. */
    static string_t toStringOf(IntValue* p, bool readable);
};

/** Integral value. This represents an integer or boolean value. Those behave identical
    in all operations. We track whether it's an integer or boolean to support meaningful
    stringification only. This is done in a derived class to keep the total size low.
    This way, an IntIntValue takes two words (8 bytes). With an additional 'is_bool'
    member, as used in PCC < 1.1.19, it takes 12 bytes. */
class IntIntValue : public IntValue {
    int32_t value;
 public:
    IntIntValue(int32_t value);
    ~IntIntValue();

    /** Get integer value. */
    int32_t getValue() const
        { return value; }

    /** Modify in-place. Use with care. */
    void add(int32_t delta)
        { value += delta; }

    /** Check whether this is a boolean value. Overloaded in derived class. */
    virtual bool isBool() const = 0;
};

/** Integral value, integer type. Provides stringification and I/O for integers. */
class IntIntIntValue : public IntIntValue {
 public:
    IntIntIntValue(int32_t value)
        : IntIntValue(value)
        { }
    ~IntIntIntValue();

    string_t toString(bool readable);
    void store(IntTagNode& sv, Stream& aux);
    IntIntIntValue* clone();
    bool isBool() const;
};

/** Integral value, integer type. Provides stringification and I/O for booleans. */
class IntIntBoolValue : public IntIntValue {
 public:
    IntIntBoolValue(int32_t value)
        : IntIntValue(value)
        { }
    ~IntIntBoolValue();

    string_t toString(bool readable);
    void store(IntTagNode& sv, Stream& aux);
    IntIntBoolValue* clone();
    bool isBool() const;
};

/** String value. */
class IntStringValue : public IntValue {
    string_t value;
 public:
    IntStringValue(string_t value);
    IntStringValue(const IntTagNode& sv, Stream& aux);
    ~IntStringValue();

    string_t toString(bool readable);
    void     store(IntTagNode& sv, Stream& aux);
    IntStringValue* clone();

    /** Get string value. */
    const string_t& getValue() const
        { return value; }
};

/** Floating-Point Value. We represent floating point as double internally,
    to achieve some mentionable precision. For external storage, floating-point
    values are converted to 48-bit format (the same as used in PCC 1.x). */
class IntFloatValue : public IntValue {
    double value;
 public:
    IntFloatValue(double value);
    IntFloatValue(const IntTagNode& sv, Stream& aux);
    ~IntFloatValue();

    string_t toString(bool readable);
    void     store(IntTagNode& sv, Stream& aux);
    IntFloatValue* clone();

    double getValue() const
        { return value; }

    /** Modify in-place. Use with care. */
    void add(int32_t delta)
        { value += delta; }
};

/** Blob value. This is the buffer used for file I/O. PCC 1.x used
    strings instead. */
class IntBlobValue : public IntValue {
    std::vector<char> value;
 public:
    IntBlobValue();
    IntBlobValue(const IntTagNode& sv, Stream& aux);
    ~IntBlobValue();

    string_t toString(bool readable);
    void     store(IntTagNode& sv, Stream& aux);
    IntBlobValue* clone();

    /** Get content. For simplicity, we expose the content in
        modifyable form. Care must still be taken to never modify any
        "live" objects. This is intended to modify newly-built objects
        only. */
    std::vector<char>& getValue()
        { return value; }
};

/** File descriptor. This is a special type instead of a compiler
    feature because we occasionally hand it into regular functions,
    which are not compiled specially. It's otherwise just an
    integer. */
class IntFileValue : public IntValue {
    int32_t file_nr;
 public:
    IntFileValue(int32_t file_nr);
    IntFileValue(const IntTagNode& sv, Stream& aux);
    ~IntFileValue();

    string_t toString(bool readable);
    void     store(IntTagNode& sv, Stream& aux);
    IntFileValue* clone();

    /** Get integer value. */
    int32_t getFileNumber() const
        { return file_nr; }
};

IntValue* makeBoolValue(int value);
IntValue* makeIntValue(int32_t value);
IntValue* makeFloatValue(double value);
IntValue* makeStringValue(string_t str);
IntValue* makeStringValue(const char* str);
int       getBoolValue(IntValue* value);

string_t  quoteString(const string_t& value);

/******************************** Inlines ********************************/

inline
IntValue::~IntValue()
{ }

/** Construct from literal.
    \param value   Value */
inline
IntIntValue::IntIntValue(int32_t value)
    : value(value)
{ }

inline
IntIntValue::~IntIntValue()
{ }

#endif
