/**
  *  \file ui/layout.h
  *  \brief Layout Management
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_UI_LAYOUT_H
#define PCC_V2_UI_LAYOUT_H

#include <vector>
#include "ui/ui.h"

/** \defgroup layout_manager Layout Management
    Classes to arrange a widget container's children. */
//@{

/** Basic Layout Manager Interface. Encapsulates the functionality
    of arranging a widget container's children. */
class UILayoutManager : public Object {
 public:
    /** Arrange a container's children. This does not resize
        the container itself. */
    virtual void doLayout(UILayoutableWidget& container, GfxRect old_size, GfxRect new_size) = 0;
    /** Get layout information. Compute preferred and minimum sizes,
        just like UIBaseWidget::getLayoutInfo(). */
    virtual void getLayoutInfo(UILayoutableWidget& container, UIBaseWidget::LayoutInfo& info) = 0;
};

// horizontal box
class UIHBoxLayout : public UILayoutManager {
    void doLayoutWithWidths(UILayoutableWidget& container, GfxRect size, std::vector<int>& sizes, int space, int outer);
    int space, outer;
 public:
    UIHBoxLayout(int space = 0, int outer = 0);
    void doLayout(UILayoutableWidget& container, GfxRect old_size, GfxRect new_size);
    void getLayoutInfo(UILayoutableWidget& container, UIBaseWidget::LayoutInfo& info);

    static UIHBoxLayout instance0, instance5;
};

// vertical box
class UIVBoxLayout : public UILayoutManager {
    void doLayoutWithHeights(UILayoutableWidget& container, GfxRect size, std::vector<int>& sizes, int space, int outer);
    int space, outer;
 public:
    UIVBoxLayout(int space = 0, int outer = 0);
    void doLayout(UILayoutableWidget& container, GfxRect old_size, GfxRect new_size);
    void getLayoutInfo(UILayoutableWidget& container, UIBaseWidget::LayoutInfo& info);

    static UIVBoxLayout instance0, instance5;
};

// grid
class UIGridLayout : public UILayoutManager {
    int x, space, outer;
    int forced_cell_height;
    int forced_cell_width;
 public:
    UIGridLayout(int x, int space = 5, int outer = 0);
    void doLayout(UILayoutableWidget& container, GfxRect old_size, GfxRect new_size);
    void getLayoutInfo(UILayoutableWidget& container, UIBaseWidget::LayoutInfo& info);
    void setForcedCellSize(int forced_cell_width, int forced_cell_height);

    int getForcedCellWidth() const
        { return forced_cell_width; }
    int getForcedCellHeight() const
        { return forced_cell_height; }
};

class UIFlowLayout : public UILayoutManager {
    int  num_lines;
    bool right_just;
    int  hgap, vgap;
 public:
    /** Constructor.
        \param num_lines number of lines to allocate space for
        \param right_just allocate from bottom-right, not top-left
        \param hgap distance between widgets in X direction
        \param vgap distance between lines in Y direction */
    UIFlowLayout(int num_lines, bool right_just, int hgap = 5, int vgap = 5)
        : num_lines(num_lines), right_just(right_just), hgap(hgap), vgap(vgap)
        { }
    void doLayout(UILayoutableWidget& container, GfxRect old_size, GfxRect new_size);
    void getLayoutInfo(UILayoutableWidget& container, UIBaseWidget::LayoutInfo& info);
};

class UINullLayout : public UILayoutManager {
 public:
    UINullLayout();
    void doLayout(UILayoutableWidget& container, GfxRect old_size, GfxRect new_size);
    void getLayoutInfo(UILayoutableWidget& container, UIBaseWidget::LayoutInfo& info);

    static UINullLayout instance;
};

//@}

#endif
