/**
  *  \file except.h
  *  \brief Exception Classes
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_EXCEPT_H
#define PCC_V2_EXCEPT_H

#include <exception>
#include <stdexcept>
#include "cpluslib/string.h"

class Stream;

/*! \defgroup Exceptions Exceptions

    Standard exceptions, with ours added:
    <pre>exception
     bad_alloc
     bad_exception
     ios::failure
     bad_typeid
     bad_cast
     runtime_error
       FileProblemException
         FileOpenException
         FileFormatException
           FileTooShortException
         FileIOException
           FileWriteException
       GraphicsIOException
       UserException
       InitException
       range_error
       overflow_error
       underflow_error
     logic_error
       length_error
       domain_error
       out_of_range
       invalid_argument</pre> */
//@{

/** Base class for file-related problems. The message is a description
    of the problem, without trailing "." (e.g. a 'strerror' string).
    The exception includes a file name for reference by the catcher. */
class FileProblemException : public std::runtime_error {
    string_t filename;
 public:
    FileProblemException(const string_t& filename, const string_t& what);
    FileProblemException(Stream& s, const string_t& what);
    FileProblemException(const FileProblemException& e);
    ~FileProblemException() throw();
    const string_t& getFileName() const
        { return filename; }
};

/** Error opening a file. Thrown by open() functions upon failure. */
class FileOpenException : public FileProblemException {
 public:
    FileOpenException(const string_t& filename, const string_t& what);
    FileOpenException(const FileOpenException&);
    ~FileOpenException() throw();
};

/** Invalid file format. Thrown by file parsers when they detect file
    format problems. */
class FileFormatException : public FileProblemException {
 public:
    FileFormatException(const string_t& filename, const string_t& what);
    FileFormatException(Stream& s, const string_t& what);
    FileFormatException(const FileFormatException&);
    ~FileFormatException() throw();
};

/** File too short. Thrown when trying to read from a file but not
    enough data can be read. Convenience class that fills in a
    standard error message; catchers should use
    FileFormatException. */
class FileTooShortException : public FileFormatException {
 public:
    FileTooShortException(const string_t& fname);
    FileTooShortException(Stream& s);
    FileTooShortException(const char* fname);
    FileTooShortException(const FileTooShortException&);
    ~FileTooShortException() throw();
};

/** File operation failed. Thrown upon I/O errors of all kind. */
class FileIOException : public FileProblemException {
 public:
    FileIOException(const string_t& filename, const string_t& what);
    FileIOException(Stream& s, const string_t& what);
    FileIOException(const FileIOException&);
    ~FileIOException() throw();
};

/** File writing failed. Thrown when trying to write but not all data
    can be written. Convenience class that fills in a standard error
    message; catchers should use FileIOException. */
class FileWriteException : public FileIOException {
 public:
    FileWriteException(const string_t& filename);
    FileWriteException(Stream& s);
    FileWriteException(const FileWriteException&);
    ~FileWriteException() throw();
};

/** Graphics operation failed. Those are usually fatal. The message
    text usually is SDL_GetError(). */
class GraphicsIOException : public std::runtime_error {
 public:
    GraphicsIOException(const string_t& awhat);
    GraphicsIOException(const char* awhat);
    GraphicsIOException(const GraphicsIOException&);
    ~GraphicsIOException() throw();
};

/** User Notify. The message should be something that can be put in
    a message box. */
class UserException : public std::runtime_error {
 public:
    UserException(const string_t& awhat);
    UserException(const char* awhat);
    UserException(const UserException&);
    ~UserException() throw();
};

/** Initialisation exception. This one is thrown when initialisation
    of some hardware fails. */
class InitException : public std::runtime_error {
 public:
    InitException(const string_t& awhat);
    InitException(const char* awhat);
    InitException(const InitException&);
    ~InitException() throw();
};

//@}

#endif
